from fuchs import FuchsTemplate as __foxt
from ermine.response import HTMLResponse


class FoxTemplates(__foxt):
    def __init__(self, dirname: str):
        super().__init__(dirname=dirname)

    def render(self, template_name: str, **kwargs) -> bytes:
        html = super().render(template_name, **kwargs)
        return HTMLResponse(html) 





from http.cookies import SimpleCookie
import json
from urllib.parse import parse_qsl
from multidict import CIMultiDict
from typing import Optional
from ermine.enum import ConnectionType


class Request:
    """class representing a request"""

    def __init__(self, scope: dict, receive, send) -> None:
        self._receive = receive
        self._send = send
        self._scope = scope
        self._req_headers: Optional[CIMultiDict] = None
        self._req_cookies: Optional[SimpleCookie] = None
        self.http_body: bytes = b""
        self.__http_has_more_body: bool = True
        self.__http_received_body_length: int = 0

    @property
    def path(self) -> str:
        """return the path of the request"""
        return self._scope['path']

    @property
    def method(self) -> str:
        """return the method of the request"""
        return self._scope['method'].lower()

    @property
    def headers(self) -> CIMultiDict:
        """return the headers of the request"""
        if not self._req_headers:
            self._req_headers = CIMultiDict(
                [(k.decode("ascii"), v.decode("ascii")) for (k, v) in self._scope["headers"]])
        return self._req_headers

    @property
    def client(self) -> str:
        """return the client of the request"""
        return self._scope['client']

    @property
    def cookies_raw(self) -> SimpleCookie:
        """return the raw cookies of the request"""
        if not self._req_headers:
            self._req_headers = CIMultiDict(
                [(k.decode("ascii"), v.decode("ascii")) for (k, v) in self._scope["headers"]])
            self._req_cookies.load(self._req_headers.get("cookie", {}))
        return self._req_cookies

    @property
    def cookies(self) -> dict:
        """return the cookies of the request"""
        return {key: m.value for key, m in self.cookies_raw.items()}

    @property
    def scope(self) -> dict:
        """return the scope of the request"""
        return self._scope

    @property
    def query(self) -> dict:
        """return the query of the request"""
        return CIMultiDict(parse_qsl(self.scope.get("query_string", b"").decode("utf-8")))

    @property
    def type(self) -> str:
        """return the type of the request"""
        return ConnectionType.ws if self.scope.get("type") == "websocket" else ConnectionType.http

    async def handle(self, message) -> None:
        if message.get("type") == "http.disconnect":
            raise Exception("Disconnect")

    async def __body_iter(self):
        if not self.type == ConnectionType.http:
            raise Exception("Not an HTTP connection")
        if self.__http_received_body_length > 0 and self.__http_has_more_body:
            raise Exception("body iter is already started and is not finished")
        if self.__http_received_body_length > 0 and not self.__http_has_more_body:
            yield self.http_body

        req_body_length: int | None = (int(self.headers.get("content-length", "0"))
                                       if not self.headers.get("transfer-encoding") == "chunked"
                                       else None)
        
        while self.__http_has_more_body:
            if req_body_length and self.__http_received_body_length > req_body_length:
                raise Exception("body length exceeded")

            message = await self._receive()
            message_type: str = message.get("type") 
            await self.handle(message)
            if message_type != "http.request":
                continue
            chunk: bytes = message.get("body", b"")
            if not isinstance(chunk, bytes):
                raise RuntimeError("Chunk is not bytes")
            self.http_body += chunk
            self.__http_has_more_body = message.get("more_body", False)
            self.__http_received_body_length += len(chunk)
            yield bytes(chunk)
    
    async def body(self) -> bytes | dict:
        """return the body of the request"""
        data: bytes = b"".join([chunk async for chunk in self.__body_iter()])
        try: 
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            return data


from http.cookies import BaseCookie, SimpleCookie
import json
import typing


class Response:
    """
    The Response class is used to return data from a request.
    """

    media_type: str | None = None
    charset: str = "utf-8"

    def __init__(
        self,
        content: typing.Any,
        status: int = 200,
        headers: dict | None = None,
        media_type: str | None = None,
    ) -> None:
        self.status: int = status
        self.content: typing.Any = content
        if media_type:
            self.media_type = media_type
        self.body = self.render(content)
        self.set_headers(headers)

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""

        if isinstance(content, bytes):
            return content

        return content.encode(self.charset)

    def set_headers(self, headers: None | typing.Mapping[str, str] = None) -> None:
        if headers is None:
            raw_headers: typing.List[typing.Tuple[bytes, bytes]] = []
        else:
            raw_headers = [
                (k.encode(self.charset), v.encode(self.charset))
                for k, v in headers.items()
            ]

        body: bytes = getattr(self, "body", b"")
        if body:
            raw_headers.append(("Content-Length", str(len(body)).encode(self.charset)))
        ctype: str = self.media_type
        if ctype:
            if ctype.startswith("text/"):
                ctype += ";" + "charset=" + self.charset + ";"
            raw_headers.append((b"content-type", ctype.encode(self.charset)))
        self.headers = raw_headers

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int = None,
        expires: int = None,
        path: str = "/",
        domain: str = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: str = "lax",
    ) -> None:
        """Adds a cookie to the answer"""
        cookie: BaseCookie = SimpleCookie()
        cookie[key] = value
        if max_age:
            cookie[key]["max-age"] = max_age
        if expires:
            cookie[key]["expires"] = expires
        if path:
            cookie[key]["path"] = path
        if domain:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.headers.append((b"set-cookie", cookie_val.encode("latin-1")))

    async def __call__(self, scope, receive, send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status,
                "headers": self.headers,
            }
        )
        await send({"type": "http.response.body", "body": self.body})


class HTMLResponse(Response):
    media_type = "text/html"

class JsonResponse(Response):
    media_type = "application/json"


    def render(self, content: dict) -> bytes:
        return json.dumps(content, indent=None).encode("utf-8")


class TextResponse(Response):
    media_type = "text/plain"

    


import inspect
import typing


class Inject:
    """
    Injectable class.
    """
    def __init__(self, injection) -> None:
        self.injection = injection


    async def __call__(self, **kwargs) -> typing.Any:
        if inspect.iscoroutinefunction(self.dep):
            return await self.dep(**kwargs)
        else:
            return self.dep(**kwargs)
class Depends:
    """
    Dependency class.
    """
    def __init__(self, dep) -> None:
        self.dep = dep


    async def __call__(self, **kwargs) -> bool:
        if inspect.iscoroutinefunction(self.dep):
            return await self.dep(**kwargs)
        else:
            return self.dep(**kwargs)
  


from ermine.groups import Group
from ermine.plugs import Pluggable
from ermine.request import Request
from ermine.plugs.alternator import Alternator
from ermine.plugs.event import Eventlistener
from ermine.static import StaticFiles
from radish import Radish
import traceback
import typing


class Ermine:
    def __init__(
        self,
        title: str = "ermine",
        description: str = "",
        redirect_slashes: bool = True,
    ) -> None:
        self.title: str = title
        self.description: str = description
        #
        self.__router = Radish(trim_trailing_slash=redirect_slashes)
        self.__alternator = Alternator()
        self.__event_listener = Eventlistener()

    async def __call__(self, scope: dict, receive, send) -> None:
        try:
            if scope["type"] == "lifespan":
                message = await receive()
                if message["type"] == "lifespan.startup":
                    await self.__event_listener("startup")
                    await self.__event_listener(message["type"])
                    await send({"type": "lifespan.startup.complete"})
                elif message["type"] == "lifespan.shutdown":
                    await self.__event_listener("shutdown")
                    await self.__event_listener(message["type"])
                    await send({"type": "lifespan.shutdown.complete"})
                    return

            elif scope["type"] == "http":
                req = Request(scope, receive, send)
                route = self.__router.get(req.method, req.path)
                resp = await self.__alternator(req, route["handler"], route["params"])
                await resp(scope, receive, send)

        except:
            traceback.print_exc()
            await send({"type": "http.response.start", "status": 500})
            await send({"type": "http.response.body", "body": b"Internal Server Error"})

    def __event_listener(self, event: str) -> None:
        # print(f"Event: {event}")
        pass

    def get(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.__router.insert("get", path, handler,)
            return handler

        return wrapper

    def post(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.__router.insert("post", path, handler)
            return handler

        return wrapper

    def put(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.__router.insert("put", path, handler)
            return handler

        return wrapper

    def delete(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.__router.insert("delete", path, handler)
            return handler

        return wrapper

    def route(
        self, path: str, dependencies: list = [], methods: list | tuple = ("get",)
    ):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            for method in methods:
                self.__router.insert(method, path, handler)
            return handler
        return wrapper

    def mount(self, plugin: Pluggable, prefix: str = None) -> None:
        if prefix and not prefix.startswith("/"):
            raise Exception("Prefix must start with '/'")

        if isinstance(plugin, Group):
            prefix = plugin.prefix or prefix
            if not prefix:
                raise Exception("Prefix cannot be empty")
            for path, func, method, deps in plugin.routes:
                if not self.__router.strict_mode and path.endswith("/"):
                    path = path[:-1]
                self.__router.add(f"{prefix}{path}", func, method, deps)
        if isinstance(plugin, StaticFiles):
            if not prefix:
                raise Exception("Prefix cannot be empty")
            self.__router.add(f"{prefix}/*filename", plugin, "get", [])

        return True

    def on(self, event: str) -> None:
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.__event_listener.add(event, handler)
            return handler

        return wrapper


from mimetypes import MimeTypes
from . import Pluggable
from ermine.response import Response

class Static(Pluggable):
    def __init__(self, path):
        self.path: str = path

    async def __call__(self, filepath: str) -> Response:
        mime = MimeTypes()
        file_type = mime.guess_type(self.path + filepath)
        file: str = self.path + ("/" + filepath) if not self.path.endswith("/") else self.path + filepath
        with open(file, "rb") as f:
            content = f.read()
        response = Response(content, headers={"content-type": file_type[0]}, media_type=file_type[0])
        return response


from . import Pluggable
import typing


class Routable(Pluggable):
    def __init__(self) -> None:
        self.routes: list = []

    def get(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.routes.append((path, handler, "get", dependencies))
            return handler

        return wrapper

    def post(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.routes.append((path, handler, "post", dependencies))
            return handler

        return wrapper

    def put(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.routes.append((path, handler, "put", dependencies))
            return handler

        return wrapper

    def delete(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.routes.append((path, handler, "delete", dependencies))
            return handler

        return wrapper

    def mount(self, plug: Pluggable, prefix: str = None) -> bool:
        # if isinstance(plug, StaticFiles):
        #     self.routes.append((prefix + "/*filename", plug, "get"))
        #     return True
        if isinstance(plug, type(self)):
            for route in plug.routes:
                self.routes.append((prefix + route[0], route[1], route[2], route[3]))
            return True



from . import Pluggable
import typing
import inspect

class Event(typing.NamedTuple):
    event: str
    handler: typing.Callable

class Eventlistener(Pluggable):

    def __init__(self) -> None:
        super().__init__()
        self.events: typing.List[Event] = []

    async def __call__(self, event: str,*args, **kwargs) -> None:
        for ev in self.events:
            if ev.event == event:
                if inspect.iscoroutinefunction(ev.handler):
                    await ev.handler(*args, **kwargs)
                else:
                    ev.handler(*args, **kwargs)


    def add(self, event: str, handler: typing.Callable):
        self.events.append(Event(event, handler))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ermine', 'ermine.plugs']

package_data = \
{'': ['*']}

install_requires = \
['multidict>=6.0.2,<7.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'radish-router>=0.1.1,<0.2.0',
 'wire-fuchs>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'ermine',
    'version': '0.4.4',
    'description': 'Nimble trough the web',
    'long_description': '<h1 align="center">Ermine</h1>\n<p align="center"> Easy, fast, stable </p>\n<img src="https://cdn.discordapp.com/attachments/857979752991031296/942470898286485524/Logo1.svg" align="right" style="margin-top: -50px;"/>\n<br>\n<br>\nErmine is designed to provide the user with the greatest possible comfort when creating Rest APIs or entire websites.\nEverything is simple and, above all, intuitively designed. No focus on superfluous configurations. Everything works, simply.\n\n🔑 Key features\n\n- intuitive, due to the clear design\n- simple, due to the fast learning curve\n- practical, through the great editor support\n- minimalistic, no superfluous functions\n\n#### What is Ermine and what is not\n\n\nErmine is not a HighSpeed framework. Ermine is probably not ready for production. Ermine is a spare time project of mine. Ermine is self-contained. It doesn\'t need anything, except for an ASGI server. So it\'s like Starlette.\nI would appreciate if you use Ermine, try it and give me your feedback.\n\n#### Participate in Ermine\n\nYou are welcome to collaborate on Ermine. However, you should maintain the codestyle, and also follow PEP 8 (the Python style guide).\n\n#### Ermine disadvantages\n\nErmine is still deep in development, which is why some features are still missing. \n\n- Websockets\n\n#### Examples\n\nHere is the most basic example of ermine\n\n```py\nfrom ermine import Ermine, Request\n\napp = Ermine()\n\n@app.get("/home")\nasync def home():\n\treturn "Welcome home"\n```\n\nYou want to build a RestAPI? No problem\n\n```py\nfrom ermine import Ermine, Request\n\n\napp = Ermine()\ntemplates = FoxTemplates("templates")\n\n@app.get("/api")\ndef api():\n\treturn {"name": "Leo", "age": 16}\n```\n\nYou want to send HTML files? Ermine got your back\n\n```py\nfrom ermine import Ermine, Request\nfrom ermine.responses import HTMLResponse\n\n\napp = Ermine()\n\n@app.get("/html")\nasync def home():\n\twith open("home.html", "r") as f:\n\t\tdata = f.read()\n\treturn HTMLResponse(data)\n```\n\nYou want to use some templates ? You want to load templates? No problem with [Fuchs](https://github.com/cheetahbyte/fuchs)\n\n```py\nfrom ermine import Ermine, Request\nfrom ermine.templating import FoxTemplates\n\napp = Ermine()\ntemplates = FoxTemplates("templates")\n\n@app.get("/home")\nasync def home():\n\treturn templates.render("home.html", name="Leo")\n```\n\n**Changes incoming**\n<center>\n\nJoin our [discord](https://discord.gg/EtqGfBVuZS) !\n\n<img src="https://images.unsplash.com/photo-1548714859-18c34a4c384a?ixlib=rb-1.2.1&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=1074&q=80" alt="Ermine" style="height: 300px" /></center>\n',
    'author': 'cheetahbyte',
    'author_email': 'bernerdoodle@outlook.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from ermine.injections import Depends, Inject
from ermine.response import Response, JsonResponse, TextResponse
from ermine.request import Request
from ermine.static import StaticFiles
from . import Pluggable
import inspect
import typing
import pydantic
import dataclasses

class Alternator(Pluggable):
    async def __call__(
        self,
        req: Request,
        func: typing.Callable[[], typing.Any],
        params: dict,
        expect_resp: bool = True
    ) -> Response:
        # if function is none, return
        if func is None:
            return TextResponse("Not Found", 404)
        

        # arguments
        arguments: dict = dict()
        # get args of function
        args = [
            (name, typ)
            for name, typ in list(inspect.signature(func).parameters.items())
        ]

        # replace arg with request object, if arg[0][1] is instance of Request
        for arg in args:
            if arg[1].annotation == Request:
                arguments[arg[0]] = req
            elif isinstance(arg[1].default, Inject):
                dependency = await self.__call__(req, arg[1].default.injection, {}, [], expect_resp=False)
                if dependency is None:
                    return TextResponse("Failed check", 500)
                arguments[arg[0]] = dependency
            elif isinstance(arg[1].annotation, pydantic.BaseModel):
                cls = arg[1]
                arguments[arg[0]] = cls(**(await req.body()))
            elif dataclasses.is_dataclass(arg[1].annotation):
                cls = arg[1].annotation
                kwargs: dict = await req.body()
                arguments[arg[0]] = cls(**kwargs)
            else:
                if arg[0] in req.query.keys():
                    arguments[arg[0]] = req.query[arg[0]]       
            val = arguments.get(arg[0])
            if not val:
                val = arg[1].default
        # update arguments with params
        arguments.update(params)

        # for dependency in deps:
        #     assert isinstance(dependency, Depends)
        #     dependency: Depends = dependency
        #     va: bool = await self.__call__(req, dependency.dep, {}, [], expect_resp=False)
        #     if va:
        #         continue
        #     else:
        #         return TextResponse("Failed check", 500)
        if inspect.iscoroutinefunction(func):
            response: typing.Union[Response, typing.Any] = await func(**arguments)
            if not expect_resp:
                return response
            if not isinstance(response, Response):
                if type(response) == str:
                    response = TextResponse(response, 200)
                elif type(response) == dict:
                    response = JsonResponse(response, 200)
            if isinstance(response, pydantic.BaseModel):
                response = JsonResponse(response.dict())
            if dataclasses.is_dataclass(response):
                response = JsonResponse(dataclasses.asdict(response))
            return response

        # static files
        elif isinstance(func, StaticFiles):
            resp: Response = await func(arguments["filename"])
            return resp

        # sync function
        else:
            response: typing.Union[Response, typing.Any] = func(**arguments)
            if not expect_resp:
                return response
            if not isinstance(response, Response):
                if type(response) == str:
                    response = TextResponse(response, 200)
                elif type(response) == dict:
                    response = JsonResponse(response, 200)
            if isinstance(response, pydantic.BaseModel):
                response = JsonResponse(response.dict())
            if dataclasses.is_dataclass(response):
                response = JsonResponse(dataclasses.asdict(response))
            return response